﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PokTiles
{
    public static class Exporter
    {
        public static string ExportStructures()
        {
            StringBuilder sb = new StringBuilder();

            //Headers
            sb.AppendLine("#include <stdint.h>");
            sb.AppendLine("#include <bitset>");

            sb.AppendLine(@"
typedef struct
{
    uint16_t Columns;
    uint16_t Rows;
    uint8_t CellWidth;
    uint8_t CellHeight;
    uint16_t PixelWidth;
    uint16_t PixelHeight;
    uint16_t DamageTilesCount;
    uint16_t AnimationTilesCount;
    uint16_t ActionTilesCount;
}MapConfig;");

            sb.AppendLine(@"
typedef struct
{
    uint8_t Column;
    uint8_t Row;
    uint8_t Life;
}TileDamage;");
            sb.AppendLine();

            sb.AppendLine(@"
typedef struct
{
    uint8_t Sprite;
    uint8_t Frame;
    uint8_t MaxFrame;
    uint8_t SpriteAnimated;
}TileAnimation;");
            sb.AppendLine();

            sb.AppendLine(@"
typedef struct
{
    uint8_t Column;
    uint8_t Row;
    uint8_t ActionType;
    uint8_t ActionParameter;
}TileAction;");
            sb.AppendLine();

            sb.AppendLine(@"
typedef struct
{
    uint8_t Sprite;
    std::bitset<8> Properties;
}Tile;");
            sb.AppendLine();

            return sb.ToString();

        }


        public static string ExportLayerConfigToC(SpriteSheet ss, Layer layer)
        {
            StringBuilder sb = new StringBuilder();

            //Count damage tiles
            sb.Append("const MapConfig layer_config_" + layer.Name + "={" +
                layer.Columns + "," +
                layer.Rows + "," +
                ss.SpriteWidth + "," +
                ss.SpriteHeight + "," +
                ss.Columns * ss.SpriteWidth + "," +
                ss.Rows * ss.SpriteHeight + "," +
                ExtractDestructibleTiles(layer).Count + "," +
                countAnimatedTileDistinct(layer) + "," +
                ExtractAcytionTiles(layer).Count  +
                "};");
            sb.AppendLine();
            return sb.ToString();
        }


        private static List<Tile> UsedTiles(SpriteSheet ss)
        {
            List<Tile> usedTiles = new List<Tile>();
            foreach (var l in ss.Layers)
            {
                foreach (var t in l.GetAllTiles())
                {
                    usedTiles.Add(t);
                }
            }
            usedTiles = usedTiles.DistinctBy(t => t.Id).ToList();
            usedTiles = usedTiles.Where(t => t.Sprite != null).ToList();
            usedTiles = usedTiles.OrderBy(t => t.Id).ToList();

            return usedTiles;
        }

        public static List<Color> ExtractPalette(int NUM_BIT, SpriteSheet ss, bool shaded = false)
        {
            var tiles = UsedTiles(ss);
            //Build a big image with all tiles to extract palette
            Bitmap tmpSpriteSheet = new Bitmap(tiles[0].Sprite.Width, tiles.Count() * tiles[0].Sprite.Height);
            Graphics g = Graphics.FromImage(tmpSpriteSheet);
            int i = 0;
            foreach (var tile in tiles)
            {
                if (tile.Sprite != null)
                {
                    g.DrawImage(tile.Sprite, 0, i * tile.Sprite.Height);
                    i++;
                }
            }
            g.Flush();

            var palette = PokittoImageConverter.ExtractFullPalette(tmpSpriteSheet, true, NUM_BIT, shaded);
            return palette;
        }


        public static string ExportSpritesToC(SpriteSheet ss, int NUM_BIT)
        {
            var palette = ExtractPalette(NUM_BIT, ss);
            var paletteShaded = ExtractPalette(NUM_BIT, ss, true);
            string res = PokittoImageConverter.ConvertPaletteToPokitto("pal", paletteShaded);
            res += "\r\n";

            var usedTiles = UsedTiles(ss);
            int size = 2 + (ss.SpriteHeight * (ss.SpriteWidth / (8 / NUM_BIT)));
            res += String.Format("const uint8_t sprites[][{0}] =\r\n{{", size);
            foreach (var tile in usedTiles)
            {
                if (tile.Sprite != null)
                    res += PokittoImageConverter.ConvertTileImageToPokitto(tile, NUM_BIT, palette);
            }
            res += "};\r\n";
            return res;
        }

        public static List<Tile> ExtractDestructibleTiles(Layer layer)
        {
            //Get all destructibles tiles
            var destructibles = new List<Tile>();
            foreach (var t in layer.GetAllTiles())
            {
                if (t.Destructible)
                    destructibles.Add(t);
            }
            return destructibles;
        }

        public static string ExportDestructibleTilesToC(Layer layer)
        {
            //Get all destructibles tiles
            var destructibles = ExtractDestructibleTiles(layer);

            StringBuilder sb = new StringBuilder();
            sb.AppendFormat("TileDamage layer_damage_{0}[] =", layer.Name);
            sb.AppendLine();
            sb.AppendLine("{");

            for (int y = 0; y < layer.Rows; y++)
            {
                for (int x = 0; x < layer.Columns; x++)
                {
                    if (layer.GetTile(x, y).Destructible)
                    {
                        sb.Append("{");
                        sb.AppendFormat("{0},{1},{2}", x, y, 100);
                        sb.AppendLine("},");
                    }
                }
            }

            sb.AppendLine("};");

            return sb.ToString();
        }

        private static List<Tile> ExtractAnimatedTiles(Layer layer)
        {
            //Get all aniamted tiles
            var animated = new List<Tile>();
            foreach (var t in layer.GetAllTiles())
            {
                if (t.Animated)
                    animated.Add(t);
            }
            return animated.DistinctBy(a => a.Id).ToList();
        }

        private static int countAnimatedTileDistinct(Layer layer)
        {
            var animated = ExtractAnimatedTiles(layer);
            int count = 0;
            foreach (var a in animated)
            {
                count += a.Animation.Frames.Count();
            }

            return count;
        }

        private static string ExportAnimatedTilesToC(Layer layer)
        {
            //Get all animated tiles
            var animated = ExtractAnimatedTiles(layer);

            StringBuilder sb = new StringBuilder();
            sb.AppendFormat("const TileAnimation layer_animations_{0}[] =", layer.Name);
            sb.AppendLine();
            sb.AppendLine("{");
            foreach (var a in animated)
            {
                int frame = 0;
                foreach (var ta in a.Animation.Frames)
                {
                    sb.Append("{");
                    sb.AppendFormat("{0},{1},{2},{3}", a.Id, frame, a.Animation.Frames.Count, ta.Id);
                    sb.AppendLine("},");
                    frame++;
                }
            }
            sb.AppendLine("};");
            return sb.ToString();
        }

        public static List<Tile> ExtractAcytionTiles(Layer layer)
        {
            //Get all destructibles tiles
            var destructibles = new List<Tile>();
            foreach (var t in layer.GetAllTiles())
            {
                if (t.Actions != null)
                    destructibles.Add(t);
            }
            return destructibles;
        }

        private static string ExportActionTilesToC(Layer layer)
        {
            //Get all animated tiles
            var actions = ExtractAcytionTiles(layer);

            StringBuilder sb = new StringBuilder();
            sb.AppendFormat("const TileAction layer_actions_{0}[] =", layer.Name);
            sb.AppendLine();
            sb.AppendLine("{");

            for (int y = 0; y < layer.Rows; y++)
            {
                for (int x = 0; x < layer.Columns; x++)
                {
                    var action = layer.GetTile(x, y).Actions;
                    if (action != null)
                    {
                        sb.Append("{");
                        sb.AppendFormat("{0},{1},{2},{3}", x, y, (int)action.actionType, (int)action.Parameter);
                        sb.AppendLine("},");
                    }
                }
            }

            sb.AppendLine("};");
            return sb.ToString();
        }

        public static string ExportLevelToC(SpriteSheet ss)
        {
            StringBuilder sb = new StringBuilder();
            foreach (var layer in ss.Layers)
            {
                if (layer.Visible)
                {
                    sb.AppendLine(ExportLayerToC(ss, layer));
                }
            }
            return sb.ToString();
        }

        public static string ExportLayerToC(SpriteSheet ss, Layer layer)
        {
            StringBuilder sb = new StringBuilder();

            //Export configuration
            sb.Append(Exporter.ExportLayerConfigToC(ss, layer));
            //Export layer
            sb.AppendFormat("const Tile layer_{0}_{1}[{2}][{3}]=", ss.Name, layer.Name, layer.Rows, layer.Columns);
            sb.AppendLine();
            sb.AppendLine("{");
            for (int y = 0; y < layer.Rows; y++)
            {
                StringBuilder sbLine = new StringBuilder();
                sbLine.Append("{");
                for (int x = 0; x < layer.Columns; x++)
                {
                    StringBuilder sbTile = new StringBuilder();
                    var tile = layer.GetTile(x, y);

                    string bitsProp = (tile.FlipH ? "1" : "0") +
                        (tile.FlipV ? "1" : "0") +
                        (tile.Animated ? "1" : "0") +
                        (tile.Solid ? "1" : "0") +
                        (tile.OneWay ? "1" : "0") +
                        (tile.Destructible ? "1" : "0") +
                        (tile.Actioned ? "1" : "0") +
                        (tile.Damaging ? "1" : "0");

                    var sBitsProp = "0b" + Utility.ReverseString(bitsProp);

                    sbTile.Append("{" + tile.Id.ToString().PadLeft(3, ' ') + "," + sBitsProp + "}");
                    //Append tile to the line
                    sbLine.Append(sbTile.ToString() + ",");
                }
                sbLine = sbLine.Remove(sbLine.Length - 1, 1);
                sb.Append(sbLine.ToString());
                sb.AppendLine("},");
            }
            sb.AppendLine("};");

            sb.AppendLine();
            sb.Append(ExportDestructibleTilesToC(layer));
            sb.AppendLine();
            sb.Append(ExportAnimatedTilesToC(layer));
            sb.AppendLine();
            sb.Append(ExportActionTilesToC(layer));
            sb.AppendLine();

            return sb.ToString();
        }


    }
}
