﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace PokTiles
{
    public partial class FormMain : Form
    {
        const int NUMBIT = 8;

        public FormMain()
        {
            InitializeComponent();
            RestoreLastJob();
        }

        private void RestoreLastJob()
        {
            //Restore last job
            LoadSpriteSheet(AppSettings.FILENAME_LAST_PALETTE);
            LoadMap(AppSettings.FILENAME_LAST_MAP);
        }

        private SpriteSheet LoadMap(string mapFileName)
        {
            var map = SpriteSheet.LoadJson(mapFileName);
            ucTilesMap1.SetSpriteSheet(map);

            SetNumericUpDownValue(numMapColumns, map.Columns);
            SetNumericUpDownValue(numMapRows, map.Rows);
            SetNumericUpDownValue(numLayerColumns, map.GetSelectedLayer().Columns);
            SetNumericUpDownValue(numLayerRows, map.GetSelectedLayer().Rows);

            bindLayersToCb();
            clbLayers.SelectedIndex = map.SelectedLayerIndex;
            return map;
        }

        private SpriteSheet LoadSpriteSheet(string paletteFileName)
        {
            var ss = SpriteSheet.LoadJson(paletteFileName);
            ucTilesMapSpriteSheet.SetSpriteSheet(ss);
            AppSettings.Singleton().LoadedSpriteSheet = ss;

            //Populate palette
            lbPalette.Items.Clear();
            foreach (var c in Exporter.ExtractPalette(NUMBIT, ss))
            {
                lbPalette.Items.Add(c);
            }

            return ss;
        }

        private void SavePaletteAs(string paleteFileName)
        {
            var ssPalette = ucTilesMapSpriteSheet.GetSpriteSheet();
            ssPalette.SaveJson(paleteFileName);
        }

        private void SaveMapAs(string mapFileName)
        {
            var ssMap = ucTilesMap1.GetSpriteSheet();
            ssMap.SaveJson(mapFileName);
        }

        private void btnNewMap_Click(object sender, EventArgs e)
        {
            LoadMap("");
        }

        private void bindLayersToCb()
        {
            var map = ucTilesMap1.GetSpriteSheet();

            clbLayers.Items.Clear();
            foreach (var l in map.Layers)
            {
                clbLayers.Items.Add(l, l.Visible);
            }
            //clbLayers.SelectedIndex = 0;
            clbLayers.DisplayMember = "Name";
            clbLayers.ValueMember = "Hide";
        }

        private void clbLayers_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            var selLayer = clbLayers.SelectedItem as Layer;
            if (selLayer != null)
                selLayer.Visible = (e.NewValue == CheckState.Checked);
            ucTilesMap1.Invalidate();
        }

        private void clbLayers_SelectedIndexChanged(object sender, EventArgs e)
        {
            int layerIndex = clbLayers.SelectedIndex;
            ucTilesMap1.SelecteLayerIndex(layerIndex);

            SetNumericUpDownValue(numLayerColumns, ucTilesMap1.GetLayerSelected().Columns);
            SetNumericUpDownValue(numLayerRows, ucTilesMap1.GetLayerSelected().Rows);
        }

        public static void SetNumericUpDownValue(NumericUpDown control, decimal value)
        {
            if (control == null) throw new ArgumentNullException(nameof(control));
            var currentValueField = control.GetType().GetField("currentValue", BindingFlags.Instance | BindingFlags.NonPublic);
            if (currentValueField != null)
            {
                currentValueField.SetValue(control, value);
                control.Text = value.ToString();
            }
        }



        private void btnSaveAsLastJob_Click(object sender, EventArgs e)
        {
            SavePaletteAs(AppSettings.FILENAME_LAST_PALETTE);
            SaveMapAs(AppSettings.FILENAME_LAST_MAP);

        }

        private void btnSaveMap_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog sfd = new SaveFileDialog())
            {
                sfd.Filter = "Pokitto Map|*.pokmap";
                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    SaveMapAs(sfd.FileName);
                }
            }

        }

        private void btnLoadMap_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog ofd = new OpenFileDialog())
            {
                ofd.Filter = "Pokitto Map|*.pokmap";
                if (ofd.ShowDialog() == DialogResult.OK)
                {
                    LoadMap(ofd.FileName);
                }
            }
        }

        private void btnLoadSpriteSheet_Click(object sender, EventArgs e)
        {
            using (FormLoadSpriteSheet flt = new FormLoadSpriteSheet())
            {
                if (flt.ShowDialog() == DialogResult.OK)
                {
                    SpriteSheet ss = flt.GetSpriteSheet();
                    AppSettings.Singleton().LoadedSpriteSheet = ss;
                    ucTilesMap1.SetTransparentColor(ss.TransparentColor);
                    ucTilesMapSpriteSheet.SetSpriteSheet(ss);
                }
            }
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            AppSettings.Save();
        }

        private void numMapColumns_ValueChanged(object sender, EventArgs e)
        {
            ucTilesMap1.SetColumns((int)numMapColumns.Value);
            SetNumericUpDownValue(numLayerColumns, ucTilesMap1.GetLayerSelected().Columns);
            SetNumericUpDownValue(numLayerRows, ucTilesMap1.GetLayerSelected().Rows);
        }

        private void numMapRows_ValueChanged(object sender, EventArgs e)
        {
            ucTilesMap1.SetRows((int)numMapRows.Value);
            SetNumericUpDownValue(numLayerColumns, ucTilesMap1.GetLayerSelected().Columns);
            SetNumericUpDownValue(numLayerRows, ucTilesMap1.GetLayerSelected().Rows);
        }

        private void numLayerColumns_ValueChanged(object sender, EventArgs e)
        {
            var layerSelected = ucTilesMap1.GetLayerSelected();
            layerSelected.Resize((int)numLayerColumns.Value, (int)numLayerRows.Value);
        }

        private void numLayerRows_ValueChanged(object sender, EventArgs e)
        {
            var layerSelected = ucTilesMap1.GetLayerSelected();
            layerSelected.Resize((int)numLayerColumns.Value, (int)numLayerRows.Value);
        }

        private void ucTilesMapSpriteSheet_TileSelected(object sender, Tile selectedTile, EventArgs e)
        {
            ucTilesMap1.PaletteTile = Utility.DeepClone(selectedTile);
            propertyGridTile.SelectedObject = selectedTile;
            ucTilesMap1.Select();
            ucTilesMap1.Invalidate();
        }

        private void btnErase_Click(object sender, EventArgs e)
        {
            ucTilesMap1.PaletteTile = new Tile();
            ucTilesMap1.Select();
            ucTilesMap1.Invalidate();
        }

        private void btnApplyAll_Click(object sender, EventArgs e)
        {
            var dr = MessageBox.Show("Apply this properties to all?", "Apply?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            if (dr == DialogResult.Yes)
            {
                var sourceTile = propertyGridTile.SelectedObject as Tile;
                var selectedLayer = ucTilesMap1.GetLayerSelected();
                if (sourceTile != null && selectedLayer != null)
                {
                    foreach (var tile in selectedLayer.GetAllTiles())
                    {
                        if(tile.Id == sourceTile.Id)
                        {
                            tile.Animated = sourceTile.Animated;

                            tile.Destructible = sourceTile.Destructible;
                            tile.FlipH = sourceTile.FlipH;
                            tile.FlipV = sourceTile.FlipV;
                            tile.Actioned = sourceTile.Actioned;
                            tile.OneWay = sourceTile.OneWay;
                            tile.Solid = sourceTile.Solid;
                            tile.Damaging = sourceTile.Damaging;
                            
                        }
                    }
                }
            }
        }

        private void ucTilesMap1_TileSelected(object sender, Tile selectedTiles, EventArgs e)
        {
            propertyGridTile.SelectedObject = selectedTiles;
        }

        private void propertyGridTile_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            var c = e.ChangedItem;
            ucTilesMap1.Invalidate();
        }

        private void btnExportToC_Click(object sender, EventArgs e)
        {
            var ss = ucTilesMapSpriteSheet.GetSpriteSheet();
            var ssMap = ucTilesMap1.GetSpriteSheet();

            var structs = Exporter.ExportStructures();
            string sprites = Exporter.ExportSpritesToC(ss, NUMBIT);
            string level = Exporter.ExportLevelToC(ssMap);

            Clipboard.SetText(structs + "\r\n" + sprites + "\r\n" + level);
        }



        #region Layers Menu
        private void moveUpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var si = clbLayers.SelectedIndex;
            var ss = ucTilesMap1.GetSpriteSheet();
            if (si > 0)
            {
                var swap = ss.Layers[si];
                ss.Layers[si] = ss.Layers[si - 1];
                ss.Layers[si - 1] = swap;
            }
            bindLayersToCb();
        }

        private void moveDownToolStripMenuItem_Click(object sender, EventArgs e)
        {

        }

        private void renameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var sl = clbLayers.SelectedItem as Layer;
            if (sl != null)
            {
                using (FormInputTex fit = new FormInputTex("New Name?"))
                {
                    if (fit.ShowDialog() == DialogResult.OK)
                    {
                        sl.Name = fit.Value;
                    }
                }
            }
            clbLayers.Invalidate();
        }

        private void cleanToolStripMenuItem_Click(object sender, EventArgs e)
        {

        }
        #endregion

        private void timerRefresh_Tick(object sender, EventArgs e)
        {
            ucTilesMap1.Invalidate();
        }

        private void lbPalette_DrawItem(object sender, DrawItemEventArgs e)
        {
            var c = lbPalette.Items[e.Index] as Color?;
            e.DrawBackground();

            if (c != null)
            {
                Color ic = ControlPaint.DarkDark(c.Value);
                Graphics g = e.Graphics;

                //Bakcground
                g.FillRectangle(new SolidBrush(c.Value), e.Bounds);
                //Text
                g.DrawString(e.Index + " " + c.ToString(), e.Font, new SolidBrush(ic),
                    new PointF(e.Bounds.X, e.Bounds.Y));
            }
            e.DrawFocusRectangle();

        }

        private void btnExportPalette_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog sfd = new SaveFileDialog())
            {
                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    var ss = ucTilesMapSpriteSheet.GetSpriteSheet();
                    var palette = Exporter.ExtractPalette(NUMBIT, ss);
                    Bitmap bmpP = new Bitmap(palette.Count, 1);
                    var g = Graphics.FromImage(bmpP);
                    for (int x = 0; x < palette.Count; x++)
                    {
                        g.FillRectangle(new SolidBrush(palette[x]), x, 0, 1, 1);
                    }
                    bmpP.Save(sfd.FileName, ImageFormat.Png);
                    g.Dispose();
                }
            }

        }

        private void ucTilesMap1_MouseMove(object sender, MouseEventArgs e)
        {
            tsStatusCoordinate.Text = ucTilesMap1.GetUnderCursorCoordinate().X + " "+
                ucTilesMap1.GetUnderCursorCoordinate().Y;
            tsStatusCoordinate.Invalidate();
        }
    }
}

