﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PokTiles
{
    static class PokittoImageConverter
    {
        public static List<Color> ExtractFullPalette(Bitmap bmp, bool FirstPixelTransparent, int numBit, bool getShaded = false)
        {
            List<Color> palette = new List<Color>();
            List<int> occurencies = new List<int>();
            if (bmp != null)
            {
                Bitmap bmpClone = (Bitmap)bmp.Clone();
                //First pixel as transparent color
                if (FirstPixelTransparent)
                {
                    Color tc = bmpClone.GetPixel(0, 0);
                    palette.Add(tc);
                    occurencies.Add(int.MaxValue);
                }

                for (int x = 0; x < bmpClone.Width; x++)
                {
                    for (int y = 0; y < bmpClone.Height; y++)
                    {
                        Color c = bmpClone.GetPixel(x, y);
                        if (c.A != 0)
                        {
                            if (palette.Contains(c))
                            {
                                var index = palette.IndexOf(c);
                                occurencies[index]++;
                            }
                            else
                            {
                                palette.Add(c);
                                occurencies.Add(0);
                            }
                        }
                    }
                }
                bmpClone.Dispose();
            }

            var pa = palette.ToArray();
            var oa = occurencies.ToArray();
            if (numBit == 8 && getShaded)
            {
                var shadedPalette = new Color[256];
                var basePalette = pa.Take(16).ToList();
                for (int i = 0; i < 16; i++)
                {
                    for (int bc = 0; bc < 16; bc++)
                    {
                        var cRGB = new ColorRGB(basePalette[bc]);
                        shadedPalette[((15 - i) * 16) + bc] = ColorRGB.FromHSL(cRGB.H,
                            cRGB.S,
                            cRGB.L / (i / 3.0 + 1.0f));
                    }
                }
                return shadedPalette.ToList();

            }
            return pa.Take((int)Math.Pow(2, numBit)).ToList();
        }

        public static List<Color> PaletteCorrected(List<Color> palette)
        {
            List<Color> newPalette = new List<Color>();

            for (int i = 0; i < palette.Count; i++)
            {
                //Correct Saturation and Luminance
                var cRGB = new ColorRGB(palette[i]);
                var newC = ColorRGB.FromHSL(cRGB.H, cRGB.S * 2.0f, cRGB.L * 1.4f);
                newPalette.Add(newC);

                ////Apply gamma correction
                //int R = (int)(palette[i].R * 1.5).Clamp(0, 255);
                //int G = (int)(palette[i].G * 1.5).Clamp(0, 255);
                //int B = (int)(palette[i].B * 1.5).Clamp(0, 255);
                //Color newColor = Color.FromArgb(R, G, B);
                //newPalette.Add(newColor);
            }
            return newPalette;
        }


        private static Color ConverToPalette(Color c, List<Color> palette)
        {
            int minIndx = 0;
            double minDst = double.MaxValue;
            double distance;
            for (int i = 0; i < palette.Count(); i++)
            {
                distance = ColorDst(c, palette[i]);
                if (distance < minDst)
                {
                    minDst = distance;
                    minIndx = i;
                }
            }

            //Alpha to color
            if (c.A == 0)
                return palette[0];
            else
                return palette[minIndx];
        }

        private static double ColorDst(Color a, Color b)
        {
            //double h = a.GetHue() - b.GetHue();
            //double u = a.GetSaturation() - b.GetSaturation();
            //double e = a.GetBrightness() - b.GetBrightness();
            //return (h * h + u * u + e * e);

            int dr = (a.R - b.R);
            int dg = (a.G - b.G);
            int db = (a.B - b.B);
            return (dr * dr + dg * dg + db * db);
        }

        private static ushort ColorTo565(Color c)
        {
            var r = (ushort)c.R >> 3;
            var g = (ushort)c.G >> 2;
            var b = (ushort)c.B >> 3;

            var c565 = (r << 11) | (g << 5) | b;
            return (ushort)c565;
        }

        public static string ConvertPaletteToPokitto(string name, List<Color> palette)
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine();
            sb.AppendLine(string.Format("//Total colors {0}", palette.Count()));
            sb.AppendLine("const uint16_t " + name + "[] = {");
            //Append colors in palette in 565 mode
            sb.AppendLine("#ifdef POK_SIM");
            foreach (var c in palette)
            {
                sb.Append(ColorTo565(c) + ",");
            }
            sb.AppendLine("");
            sb.AppendLine("#else");
            //Build palette correccted
            var paletteCorrected = PaletteCorrected(palette);
            foreach (var c in paletteCorrected)
            {
                sb.Append(ColorTo565(c) + ",");
            }
            sb.AppendLine("");
            sb.AppendLine("#endif");
            sb.AppendLine("};");
            sb.AppendLine();
            return sb.ToString();
        }

        public static string ConvertTileImageToPokitto(Tile tile, int numBit, List<Color> palette)
        {
            //Map image with reduced palette
            Bitmap img = (Bitmap)tile.Sprite;
            Bitmap bmpIndexed = new Bitmap(img.Width, img.Height);
            for (int x = 0; x < img.Width; x++)
            {
                for (int y = 0; y < img.Height; y++)
                {
                    Color c = img.GetPixel(x, y);
                    var newC = ConverToPalette(c, palette);
                    bmpIndexed.SetPixel(x, y, newC);
                }
            }

            //Extrac index of colors
            int[,] pixels = new int[img.Width, img.Height];
            for (int x = 0; x < img.Width; x++)
            {
                for (int y = 0; y < img.Height; y++)
                {
                    Color c = img.GetPixel(x, y);
                    //Transparent color to index 0
                    if (c.A == 0)
                    {
                        pixels[x, y] = 0;
                    }
                    else
                    {
                        if (palette.Contains(c))
                        {
                            pixels[x, y] = palette.IndexOf(c);
                        }
                        else
                        {
                            //Missong color to index 0
                            pixels[x, y] = 0;
                        }
                    }
                }
            }

            //Prepare output string
            StringBuilder sb = new StringBuilder();
            sb.AppendLine(@"//" + tile.Id);
            sb.AppendLine("{");
            sb.AppendLine(String.Format("{0},{1},", img.Width, img.Height));

            //Calc offset
            int offSet = 8 / numBit;
            //line by line
            for (int y = 0; y < img.Height; y++)
            {
                string line = "";
                for (int x = 0; x < img.Width; x += offSet)
                {
                    //sum bit color
                    int pc = 0;
                    for (int i = 0; i < offSet; i++)
                    {
                        if (x + (offSet - 1 - i) < img.Width)
                        {
                            pc += pixels[x + (offSet - 1 - i), y] << numBit * i;
                        }
                    }
                    line += pc + ",";
                }
                sb.AppendLine(line);
            }
            sb.Append("},\r\n");
            return sb.ToString();
        }
    }
}
